/******************************************************************************

    TiVo Desktop Mp3.dll Source Code and Build Process
    Copyright (c) 2003 TiVo Inc.

    This library is free software; you can redistribute it and/or modify it
    under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation; either version 2.1 of the License, or (at
    your option) any later version.

    This library is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser
    General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this library; if not, write to the Free Software Foundation,
    Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307 USA.

    April 4th, 2003

    www.tivo.com

    TiVo Inc.
    P.O. Box 2160
    Alviso, CA 95002

******************************************************************************/

////////////////////////////////////////////////////////////////////////////////
//
// StringUtils.cpp
//
// Copyright (c) 2002 TiVo Inc., all rights reserved.
//
////////////////////////////////////////////////////////////////////////////////

#include <ctype.h>
#include <dir.h>
#include "StringUtils.h"
#include "ModuleErrorUtils.h"

////////////////////////////////////////////////////////////////////////////////

namespace StringUtils
{

////////////////////////////////////////////////////////////////////////////////
std::string SafeString(LPCSTR str, bool check_lasterror)
{
    // This is a safe way to construct a std::string from a char *.
    // It will not throw an exception if the char * is NULL.

    // If str == NULL, and checl_lasterror is true then we'll check the
    // Windows last error value to see if we should throw an exception

    // Validate the input string
    if (str == NULL)
    {
        if (check_lasterror)
        {
            DWORD err = GetLastError();

            if (err != ERROR_NOT_FOUND)
            {
                // Throw for all errors except ERROR_NOT_FOUND
                ThrowLastWinError();
            }
        }

        return "";
    }

    return str;
}
////////////////////////////////////////////////////////////////////////////////
std::string ExtractFileName(const std::string &rstr)
{
    return ExtractFileName(rstr.c_str());
}
////////////////////////////////////////////////////////////////////////////////
std::string ExtractFileName(LPCSTR str)
{
    // Return a string containing the file name + ext only

    // Validate the input string
    if (str == NULL)
        return "";

    char path[MAXPATH+1];

    // We can't use fnsplit because that function fails on a file like:
    // c:\some_dir\...foo

    const char *p;

    // Look for the last backslash
    p = strrchr(str, '\\');

    if (!p)
    {
        // Didn't find any, so look for the last forward slash
        p = strrchr(str, '/');
    }

    // Didn't find any slashes, so use the whole string
    if (!p)
        p = str;
    else
    {
        // We did find it, so advance past the slash:
        p++;
    }

    // Save the filename+ext locally
    strcpy(path, p);

    return std::string (path);
}
////////////////////////////////////////////////////////////////////////////////
std::string ExtractFileExt(const std::string &rstr)
{
    return ExtractFileExt(rstr.c_str());
}
////////////////////////////////////////////////////////////////////////////////
std::string ExtractFileExt(LPCSTR str)
{
    // Return a string containing the extension only

    // Validate the input string
    if (str == NULL)
        return "";

    char ext[MAXPATH+1] = "";

    // We can't use fnsplit because that function fails on a file like:
    // c:\some_dir\...foo

    const char *p;

    // Get the name + ext first
    std::string path = ExtractFileName(str);

    // Look for the last period in the string
    p = strrchr(path.c_str(), '.');

    if (p)
    {
        // Found the last period, so assume that the extension starts here
        strcpy(ext, p);
    }

    return std::string (ext);
}
////////////////////////////////////////////////////////////////////////////////
std::string ExtractFileNameOnly(const std::string &rstr)
{
    return ExtractFileNameOnly(rstr.c_str());
}
////////////////////////////////////////////////////////////////////////////////
std::string ExtractFileNameOnly(LPCSTR str)
{
    // Return a string containing the filename only (no path, no ext)

    // Validate the input string
    if (str == NULL)
        return "";

    char name[MAXPATH+1] = "";

    // We can't use fnsplit because that function fails on a file like:
    // c:\some_dir\...foo

    char *p;

    // Get the name + ext first
    std::string path = ExtractFileName(str);

    // Copy to our buffer
    strcpy(name, path.c_str());

    // Look for the last period in the string
    p = strrchr(name, '.');

    if (p)
    {
        // Found the last period, so assume that the name stops here

        *p = '\0';
    }

    return std::string (name);
}
////////////////////////////////////////////////////////////////////////////////
std::string ExtractFilePath(const std::string &rstr)
{
    return ExtractFilePath(rstr.c_str());
}
////////////////////////////////////////////////////////////////////////////////
std::string ExtractFilePath(LPCSTR str)
{
    // Return a string containing the drive and path only (no filename or ext)

    // Validate the input string
    if (str == NULL)
        return "";

    char path[MAXPATH+1];

    // We can't use fnsplit because that function fails on a file like:
    // c:\some_dir\...foo

    strcpy(path, str);

    char *p;

    // Look for the last backslash
    p = strrchr(path, '\\');

    if (!p)
    {
        // Didn't find any, so look for the last forward slash
        p = strrchr(path, '/');
    }

    // If we found a slashes, terminate the string there.
    if (p)
    {
        *p = '\0';
    }

    return std::string (path);
}
////////////////////////////////////////////////////////////////////////////////
std::string ForceEndSlash(std::string str, bool bAdd)
{
    if (str.length())
    {
        // Get the last char in the string
        char last = str[str.length()-1];

        bool present = false;

        // See if there's a slash at the end
        if (last == '/' || last == '\\')
            present = true;

        // See if we're being asked to add a slash or remove one
        if (bAdd)
        {
            if (!present)
            {
                // There isn't, so add one.
                str += "/";
            }
        }
        else
        {
            if (present)
            {
                // There is one, so remove it
                str[str.length()-1] = '\0';
            }
        }
    }

    return str;
}
////////////////////////////////////////////////////////////////////////////////
std::string Trim(LPCSTR in)
{
    LPCSTR start;
    LPCSTR end;

    // Validate the input
    if (!in || !in[0])
        return "";

    // Skip any leading whitespace, and find the beginning of the string
    for(start = in; *start ; start++)
    {
        // Skip spaces
        if (isspace(*start))
            continue;

        // Skip non printable chars
        if (!isprint(*start))
            continue;

        // We found a valid char, so this is the start of the string
        break;
    }

    // Skip any trailing whitespace, and find the end of the string
    for(end = &in[strlen(in)]; end > start ; end--)
    {
        // Skip spaces
        if (isspace(*(end-1)))
            continue;

        // Skip non printable chars
        if (!isprint(*(end-1)))
            continue;

        // We found a valid char, so this is the end of the string
        break;
    }

    // Make sure there's something left
    if (start == end)
        return "";

    // Now we have the start and end pointers to the valid part of the
    // string, so we'll return a string containing only those chars.

    return std::string(start, end - start);
}
////////////////////////////////////////////////////////////////////////////////

} // namespace

////////////////////////////////////////////////////////////////////////////////
#ifdef USE_MAIN
#include <stdio.h>

using namespace StringUtils;

bool str_check(int line, const char *file, const char *a, const char *b)
{
    if (strcmp(a, b) != 0)
    {
        printf("str_check failure a != b on line %d of %s\n  a:%s\n  b:%s\n",
               line, file, a, b);
        return false;
    }

    return true;
}

#define STR_CHECK(a, b) str_check(__LINE__, __FILE__,a, b)
int main (void)
{
    std::string str1("c:\\path_name\\subdir_name\\file_name.extension");
    std::string str2("c:\\path_name\\subdir_name.with.embedded.periods\\file_name.ext");
    std::string str3("c:\\path_name\\subdir_name.with.embedded.periods\\file_name....ext");
    std::string str4("c:\\path_name\\.ext");
    std::string str5("....ext");
    std::string str6("c:/dir_name/...foo");
    std::string str7("    http://www.cnn.com/stuff.html?param1=1        ");

    std::string s;
    bool good = true;

    printf("\n");

    s = ExtractFilePath(str1);

    good = good & STR_CHECK(s.c_str(),"c:\\path_name\\subdir_name");

    s = ExtractFilePath(str2);

    good = good & STR_CHECK(s.c_str(),"c:\\path_name\\subdir_name.with.embedded.periods");

    s = ExtractFileName(str2);

    good = good & STR_CHECK(s.c_str(),"file_name.ext");

    s = ExtractFileName(str3);

    good = good & STR_CHECK(s.c_str(),"file_name....ext");

    s = ExtractFileExt(str3);

    good = good & STR_CHECK(s.c_str(),".ext");

    s = ExtractFilePath(str5);

    good = good & STR_CHECK(s.c_str(),"....ext");

    s = ExtractFileName(str6);

    good = good & STR_CHECK(s.c_str(),"...foo");

    s = Trim(str6);

    good = good & STR_CHECK(s.c_str(),"c:/dir_name/...foo");

    s = Trim(str7);

    good = good & STR_CHECK(s.c_str(),"http://www.cnn.com/stuff.html?param1=1");


    if (good)
        printf("Success.\n");
    else
        printf("Failed.\n");


    return 0;
}

#endif // USE_MAIN

