/******************************************************************************

    TiVo Desktop Mp3.dll Source Code and Build Process
    Copyright (c) 2003 TiVo Inc.

    This library is free software; you can redistribute it and/or modify it
    under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation; either version 2.1 of the License, or (at
    your option) any later version.

    This library is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser
    General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this library; if not, write to the Free Software Foundation,
    Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307 USA.

    April 4th, 2003

    www.tivo.com

    TiVo Inc.
    P.O. Box 2160
    Alviso, CA 95002

******************************************************************************/

////////////////////////////////////////////////////////////////////////////////
//
// BaseModule.h - base template for Calypso modules that derive from ModuleApi
//
//
// Copyright (c) 2002 TiVo Inc., all rights reserved.
//
////////////////////////////////////////////////////////////////////////////////

#ifndef INCLUDE_GUARD_D4186428_ED6E_4F09_9BDB_DB7596843687
#define INCLUDE_GUARD_D4186428_ED6E_4F09_9BDB_DB7596843687

#include <windows.h>
#include <string>
#include <stdarg.h>
#include <map>
#include <memory>

#include "Api/LogApi.h"
#include "Api/ModuleApi.h"
#include "Api/HostProgramApi.h"
#include "Api/DetailsApi.h"

#include "TivoApi/ApiRef.h"
#include "TivoApi/ApiHelp.h"
#include "TivoApi/Dll/DllHelp.h"
#include "Common/ApiRefs.h"
#include "Common/ModuleErrorUtils.h"
#include "Common/StringUtils.h"
#include "Common/AutoUtils.h"

using namespace TivoApi;

#ifndef TIVO_NO_LOGGING

#define TIVO_LOG(x) TheModule::GetLogApi()->Log_Append x

#else

#define TIVO_LOG(x) 0

#endif

////////////////////////////////////////////////////////////////////////////////
class BaseRoot
{
    public:

    static BaseRoot*                  m_self;
    static HostProgramApi*            m_host_program_api;

    BaseRoot()
    {
        m_self = this;
    }

    virtual ~BaseRoot()
    {
        m_self = NULL;

        if (m_host_program_api)
        {
            m_host_program_api->Api_Release();
            m_host_program_api = NULL;
        }
    }

    void InitRoot(HostProgramApi* host)
    {
        m_host_program_api = host;
        m_host_program_api->Api_AddRef();
    }

};
////////////////////////////////////////////////////////////////////////////////
template <typename T, typename T_api>
class BaseModule: public BaseRoot, public T_api
{
	public:

    typedef std::map<std::string, Api *>  SingletonApiMapType;
    typedef SingletonApiMapType::iterator SingletonApiMapIterator;

    LogRef                                m_log_api;

    SingletonApiMapType                   m_singleton_api_map;

    CRITICAL_SECTION                      m_crit;

    //-------------------------------------------------------------------------
    // Ctor
    BaseModule()
    {
        AutoCrit::Init(m_crit);
    }
    //-------------------------------------------------------------------------
    // Dtor
    ~BaseModule()
    {
        // Release and erase all the elements of the api map
        for(SingletonApiMapIterator iter=m_singleton_api_map.begin();iter!=m_singleton_api_map.end();)
        {
            // Release this api
            iter->second->Api_Release();

            // Remove it from the map
            m_singleton_api_map.erase(iter++);
        }

        // Clean up our crit
        AutoCrit::Destroy(m_crit);
    }
    //-------------------------------------------------------------------------
    // Initializes some members of this class and our base
    void InitBaseModule(HostProgramApi* host)
    {
        // Let the base initialize
        InitRoot(host);

        // Open a LogApi
        m_log_api = ApiByName("LogApi");
    }
    //-------------------------------------------------------------------------
    Api * GetBrokenDetailsApi(UrlApi *raw_url)
    {
        UrlRef url_api = raw_url;

        FilterDetailsRef broken = ApiByName("FilterDetailsApi");

        if (!broken->Filter_Initialize(NULL))
        {
            ThrowLastWinError();
        }

        // Tweak the Title and (hidden) FileName details.
        std::string full = ForceEndSlash (SafeString(url_api->Url_GetDocument()),false);

        std::string title = ExtractFileName(full);

        if (!broken->Filter_SetValue("Title", title.c_str()))
        {
            ThrowLastWinError();
        }

        if (!broken->Filter_SetValue("SourceLocation", full.c_str()))
        {
            ThrowLastWinError();
        }

        if (!broken->Filter_SetValue("FileName", title.c_str()))
        {
            ThrowLastWinError();
        }

        if (!broken->Filter_HideKey("FileName"))
        {
            ThrowLastWinError();
        }

        if (!broken->Filter_SetValue("ContentType", "x-application/tivo-broken"))
        {
            ThrowLastWinError();
        }

        if (!broken->Filter_SetValue("SourceFormat", "x-application/tivo-broken"))
        {
            ThrowLastWinError();
        }

        // Ready for use.

        return broken;
    }
    //-------------------------------------------------------------------------
    // Gets the host program api ref
    static HostProgramApi* GetHostProgramApi()
    {
        return m_host_program_api;
    }
    //-------------------------------------------------------------------------
    // Gets an api by name, keeping a copy of it and returning that same one
    // on subsequent requests.
    static Api* SingletonApiByName(LPCSTR name)
    {
        // Ensure only one thread can enter at a time
        AutoCrit ac(GetSelf()->m_crit);

        // The Api name string
        std::string str = name;

        // Look up the Api element assiciated with this name string
        Api * api = GetSelf()->m_singleton_api_map[str];

        // See if we found one
        if (!api)
        {
            // This Api does not exist yet

            // Allocate it
            api = ApiByName(name);

            // Add it to the map
            GetSelf()->m_singleton_api_map[str] = api;
        }

        // Bump the ref count and hand the Api back
        return AddRef(api);
    }
    //-------------------------------------------------------------------------
    // Gets a new api by name
    static Api* ApiByName(LPCSTR name)
    {
        // Ensure that we've been initialized
        if (!m_host_program_api)
            ThrowWinError(ERROR_INTERNAL_ERROR);

        // Allocate a new copy of the requested Api
        Api* api = m_host_program_api->Host_GetApi(name, AddRef(GetSelf()));

        // See if the allocation succeeded
        if (!api)
            ThrowLastWinError();

        // Return the new Api
        return api;
    }
    //-------------------------------------------------------------------------
    // Gets a new UrlApi and inits it to the 'url' string
    static UrlApi* NewUrl(LPCSTR url)
    {
        // Allocate a new UrlApi
        UrlRef url_api = ApiByName("UrlApi");

        // Ensure that we can initialize it to the specified url string
        if (url_api->Url_Initialize(url) == false)
            ThrowLastWinError();

        // Return the new UrlApi
        return url_api;
    }
    //-------------------------------------------------------------------------
    // Gets a new UrlBuilderApi and optionally inits it to a UrlApi
    static UrlBuilderApi* NewUrlBuilder(UrlApi* raw_url = NULL)
    {
        // Put the raw_url in a ref
        UrlRef url(raw_url);

        // Allocate a new UrlBuilderApi
        UrlBuilderRef ubld_api = ApiByName("UrlBuilderApi");

        // Ensure that we can initialize it to the specified UrlApi
        if (ubld_api->Url_Initialize(url) == false)
            ThrowLastWinError();

        // Return the new UrlBuilderApi
        return ubld_api;
    }
    //-------------------------------------------------------------------------
    // Gets the address of us as the most derived class
    static T* GetSelf()
    {
        return static_cast<T*>(m_self);
    }
    //-------------------------------------------------------------------------
    // Generic printf-like logging routine
    static void LogPrintf
    (
        DWORD   severity,
        LPCSTR  category,
        LPCSTR  format,
        ...
    )
    {
        va_list args;

        va_start(args, format);

        // Format the message.
        char *msg;
        int buflen;

        // Get the length needed to store this string
        buflen = vsnprintf(NULL, 0, format, args);

        // Allocate the string
        msg = new char[buflen+1];

        try
        {
            // Format the string
            vsprintf(msg, format, args);

            va_end(args);

            GetSelf()->m_log_api->Log_Append(severity, category, "%s", msg);
        }
        catch(...)
        {
            // Ignore any exceptions
        }

        delete [] msg;
    }
    //-------------------------------------------------------------------------
    // Specific "info" logging routine
    static void LogInfoString(LPCSTR cat, LPCSTR str)
    {
        LogPrintf(LogInfo, cat, "%s", str);
    }
    static void LogInfoString(std::string cat, std::string str)
    {
        LogPrintf(LogInfo, cat.c_str(), "%s", str.c_str());
    }
    //-------------------------------------------------------------------------
    // Specific "error" logging routines
    static void LogErrorString(LPCSTR cat, LPCSTR str)
    {
        LogPrintf(LogError, cat, "%s", str);
    }
    static void LogErrorString(std::string cat, std::string str)
    {
        LogPrintf(LogError, cat.c_str(), "%s", str.c_str());
    }
    //-------------------------------------------------------------------------
    // Gets the internal copy of the log api as a LogRef
    static LogRef GetLogApi()
    {
        return GetSelf()->m_log_api;
    }
    //-------------------------------------------------------------------------
};
////////////////////////////////////////////////////////////////////////////////
#endif // INCLUDE_GUARD