/******************************************************************************

    TiVo Desktop Mp3.dll Source Code and Build Process
    Copyright (c) 2003 TiVo Inc.

    This library is free software; you can redistribute it and/or modify it
    under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation; either version 2.1 of the License, or (at
    your option) any later version.

    This library is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser
    General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this library; if not, write to the Free Software Foundation,
    Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307 USA.

    April 4th, 2003

    www.tivo.com

    TiVo Inc.
    P.O. Box 2160
    Alviso, CA 95002

******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <Windows.h>
#include "Audio.h"
#include "Api/QualityLevel.h"
#include "Api/FileViewApi.h"

#include "AudioStream.h"

////////////////////////////////////////////////////////////////////////////////
Audio::Audio(UrlApi* raw_url) :
    m_seek_ms(0),
    m_duration_ms(-1),
    m_max_length_ms(0)
{
    m_url_api = raw_url;
    m_file_view = AudioModule::ApiByName("FileViewApi");

    if ((!m_file_view->File_Initialize(m_url_api)) || (m_file_view->Data_Get() == NULL))
        ThrowLastWinError();

    m_mpeg_audio = new MpegAudio(m_file_view->Data_GetSize(), m_file_view->Data_Get());

    if (!m_mpeg_audio->IsValidMpeg())
    {
        delete m_mpeg_audio;
        m_mpeg_audio = NULL;
        ThrowWinError(ERROR_UNSUPPORTED_TYPE);
    }

}
////////////////////////////////////////////////////////////////////////////////
Audio::~Audio()
{
    delete m_mpeg_audio;
}
////////////////////////////////////////////////////////////////////////////////
void Audio::PrepareMaxLength()
{
    if (!m_max_length_ms)
    {
        // Get the total length of this song
        m_max_length_ms = m_mpeg_audio->CalculateTotalDuration();
    }
}
////////////////////////////////////////////////////////////////////////////////
StreamApi*  TIVO_API_METHOD Audio::Audio_GetStreamApi(LPCSTR format)
{
    LONG total_length = m_file_view->Data_GetSize() - 1;
    LONG start_offset;
    LONG end_offset;
    LONG frame_len;
    int min_frames = 3;

    TIVO_API_TRY

    if (strcmp(format, "audio/mpeg") != 0)
    {
        // There's really only one mime type we can handle: "audio/mpeg" which
        // includes layers 1, 2, or 3 of either MPEG 1.0 or MPEG 2.0 (we don't
        // directly support MPEG 2.5)

        ThrowWinError(ERROR_UNSUPPORTED_TYPE);
    }

    // Check for the most common usage and optimize it
    if (m_seek_ms == 0 && m_duration_ms == -1)
    {
        // If no seeking and no duration cropping is needed, we don't need to
        // pre-scan the whole file (which PrepareMaxLength does).  Instead
        // Just tell AudioStream to use the whole file
        start_offset = 0;
        end_offset   = total_length;

    }
    else
    {
        // Ok, there is no optimization possible here.  Some combination of
        // seeking and duration cropping is needed, so we must hit the file
        // once to get the accurate duration and then hit at least part of the
        // file again to find the frame offset to which we'll seek.

        PrepareMaxLength();

        // Calculate where the starting offset is (using the seek value)
        // in the file.
        start_offset = m_mpeg_audio->CalculateOffsetAtTime
                                     (
                                        m_seek_ms,
                                        min_frames
                                     );

        // Handle errors
        if (start_offset == -1)
        {
            ThrowWinError(ERROR_INVALID_PARAMETER);
        }

        // Check for the case where the duration has not been set.  In that case
        // We assume the entire duration of the song from the seek point.
        if (m_duration_ms == -1)
        {
            end_offset = total_length;
        }
        else
        {
            // Calculate where the ending offset is (using the seek and duration
            // value) in the file.
            end_offset = m_mpeg_audio->CalculateOffsetAtTime
                                       (
                                           m_seek_ms+m_duration_ms,
                                           0
                                       );
        }

        // If the duration is too long, trim it back to the real end of the file
        if (end_offset == -1)
        {
            end_offset = total_length;
        }

        // The length (in bytes) of the last frame processed.
        frame_len = m_mpeg_audio->GetFullMp3Info()->frame_length;

        if ((end_offset - start_offset) < (min_frames * frame_len))
        {
            // Ok, the duration is too small for the mpeg players to play.
            // We must move either the start or the end pointer a bit.

            // If the end pointer isn't too close to the end, then extend the end
            // pointer.
            if (end_offset < total_length - (min_frames * frame_len))
            {
                end_offset += (min_frames * frame_len);
            }
            else
            {
                // Oops the end pointer is too close to the end.

                // First we'll set the end pointer to the very end of the image
                end_offset = total_length;

                // Then we'll back up the start pointer enough to make the distance
                // fit within the minimum.
                start_offset -= ((min_frames * frame_len) - (end_offset - start_offset));
            }
        }
    }

    return new AudioStream
               (
                   // The url of the file
                   m_url_api,

                   // Our FileView
                   m_file_view,

                   // The starting offset into the image
                   start_offset,

                   // The total size of the image that should be accessed
                   end_offset - start_offset
               );

    TIVO_API_CATCH_RETURN(NULL)
}
////////////////////////////////////////////////////////////////////////////////
BOOL          TIVO_API_METHOD Audio::Audio_Seek(DWORD time)
{
    LONG l;

    TIVO_API_TRY

    // Seeking to 0 is always acceptable
    if (time == 0)
    {
        m_seek_ms = 0;
        return true;
    }

    PrepareMaxLength();

    // Validate this seek point
    if (time > (DWORD)m_max_length_ms)
    {
        SetLastError(ERROR_INVALID_PARAMETER);
        return false;
    }

    m_seek_ms = time;

    return true;

    TIVO_API_CATCH_RETURN(false)
}
////////////////////////////////////////////////////////////////////////////////
BOOL          TIVO_API_METHOD Audio::Audio_SetDuration(DWORD time)
{
    TIVO_API_TRY

    // A Duration of 0  means the whole file.  We use -1 to indicate that
    if (time == 0)
    {
        m_duration_ms = -1;
        return true;
    }

    PrepareMaxLength();

    // Validate that this duration is less than the entire song
    if (time > (DWORD)m_max_length_ms)
    {
        SetLastError(ERROR_INVALID_PARAMETER);
        return false;
    }

    m_duration_ms = time;

    return true;

    TIVO_API_CATCH_RETURN(false)
}
////////////////////////////////////////////////////////////////////////////////
