/******************************************************************************

    TiVo Desktop Mp3.dll Source Code and Build Process
    Copyright (c) 2003 TiVo Inc.

    This library is free software; you can redistribute it and/or modify it
    under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation; either version 2.1 of the License, or (at
    your option) any later version.

    This library is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser
    General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this library; if not, write to the Free Software Foundation,
    Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307 USA.

    April 4th, 2003

    www.tivo.com

    TiVo Inc.
    P.O. Box 2160
    Alviso, CA 95002

******************************************************************************/

#include <io.h>
#include <fcntl.h>
#include <dir.h>
#include <ctype.h>
#include <windows.h>
#include <stdio.h>
#include <algorithm>
#include "AudioDetails.h"
////////////////////////////////////////////////////////////////////////////////
AudioDetails::KeyData AudioDetails::s_key_array[] =
{
    // All types
    {kdtTitle,             "Title",             true},
    {kdtType,              "ContentType",       true},
    {kdtSourceFormat,      "SourceFormat",      true},
    {kdtSourceLocation,    "SourceLocation",    true},
    {kdtFileName,          "FileName",          false}, // Hidden option
    {kdtCreationDate,      "CreationDate",      true},
    {kdtLastChangeDate,    "LastChangeDate",    true},
    {kdtLastAccessDate,    "LastAccessDate",    true},
    {kdtSourceSize,        "SourceSize",        true},
    {kdtKeywords,          "Keywords",          false},
    {kdtCopyright,         "Copyright",         false},

    // All Audio types
    {kdtSourceBitRate,     "SourceBitRate",     true},
    {kdtSourceSampleRate,  "SourceSampleRate",  true},
    {kdtMusicGenre,        "MusicGenre",        true},
    {kdtArtistName,        "ArtistName",        true},
    {kdtAlbumTitle,        "AlbumTitle",        true},
    {kdtDuration,          "Duration",          true},
    {kdtAccurateDuration,  "AccurateDuration",  false}, // Hidden option
    {kdtAlbumYear,         "AlbumYear",         true},
    {kdtRecordLabel,       "RecordLabel",       true},
    {kdtSongTitle,         "SongTitle",         true},

};
////////////////////////////////////////////////////////////////////////////////
AudioDetails::AudioDetails(UrlApi* raw_url) :
    m_key_idx(-1),
    m_accuracy(false),
    m_error(false),
    m_access_timeout(0)
{
    m_url_api = raw_url;

    m_dhelp_api = AudioModule::SingletonApiByName("DetailsHelpApi");

    m_cache_api = AudioModule::SingletonApiByName("CacheApi");
}
////////////////////////////////////////////////////////////////////////////////
AudioDetails::~AudioDetails()
{
}
////////////////////////////////////////////////////////////////////////////////
LPCSTR  TIVO_API_METHOD AudioDetails::Details_GetFirstKey()
{
    TIVO_API_TRY

    m_key_idx = 0;

    return Details_GetNextKey();

    TIVO_API_CATCH_RETURN(NULL)
}
////////////////////////////////////////////////////////////////////////////////
LPCSTR  TIVO_API_METHOD AudioDetails::Details_GetNextKey()
{
    TIVO_API_TRY

    while(m_key_idx != -1)
    {
        // Get the next values from the array
        LPCSTR str  = s_key_array[m_key_idx].key;
        bool supported = s_key_array[m_key_idx].supported;

        // Bump the index
        m_key_idx++;

        // Check for the end of the array
        if (m_key_idx >= (sizeof(s_key_array) / sizeof(s_key_array[0])))
            m_key_idx = -1;

        // Check to see if this string is supported
        if (supported)
        {
            return str;
        }
    }

    // End of the list
    SetLastError(ERROR_NO_MORE_ITEMS);
    return NULL;

    TIVO_API_CATCH_RETURN(NULL)
}
////////////////////////////////////////////////////////////////////////////////
void AudioDetails::ScanFileImage(bool full, bool mp3_or_id3)
{
    // Short circuit if we previously had an error.
    if (m_error)
        return ;

    // See if we've already gotten this info
    if (m_mp3.get() && m_id3.get())
    {
        //  Check to see if a timeout has been set
        if (m_access_timeout)
        {
            // A timeout has been set, so check to see if it has expired
            if (GetTickCount() <= m_access_timeout)
            {
                // We haven't timed out yet so we'll short circuit this call
                return ;
            }
            else
            {
                // Ok, the timeout is up so we'll reset the timeout flag and
                // try the access again.
                m_access_timeout = 0;
            }
        }
        else
        {
            // No timeout has been set so we'll short circuit this call.
            return ;
        }
    }

    FileViewRef                  file_view;
    std::auto_ptr<MpegAudio>     mpeg_audio;

    // Create a new file view to map the file image
    file_view = AudioModule::ApiByName("FileViewApi");

    // Init the file view to our document
    if (!file_view->File_Initialize(m_url_api))
    {
        DWORD err = GetLastError();

        if ((err == ERROR_FILE_NOT_FOUND) || (err == ERROR_PATH_NOT_FOUND))
        {
            // The file cannot be located, so we'll silently return after
            // setting m_mp3 and m_id3 to empty instances

            if (m_mp3.get() == NULL)
                m_mp3.reset(new Mp3Info);

            if (m_id3.get() == NULL)
                m_id3.reset(new struct id3v2);

            m_access_timeout = GetTickCount() + c_timeout_value;

            return ;
        }
        ThrowLastWinError();
    }

    // Create a new MpegAudio instance from this file view
    mpeg_audio.reset(new MpegAudio(file_view->Data_GetSize(), file_view->Data_Get()));

    // Check for validity
    if (!mpeg_audio->IsValidMpeg())
    {
        // Not valid, so we'll just return
        m_error = true;
        return;
    }

    if (mp3_or_id3)
    {
        Mp3Info *mp3;

        if (full)
        {
            // Get the accurate (but time consuming) mp3 info
            mp3 = mpeg_audio->GetFullMp3Info();
            m_accuracy = true;
        }
        else
        {
            // Get the quick (but not as accurate) mp3 info
            mp3 = mpeg_audio->GetQuickMp3Info();
            m_accuracy = false;
        }

        if (mp3)
        {
            m_mp3.reset(new Mp3Info);
            *m_mp3 = *mp3;
        }
        else
        {
            // Oops, couldn't get an Mp3Info instance
            // This should never happen!
            ThrowWinError(ERROR_INTERNAL_ERROR);
        }
    }
    else
    {
        id3v1   *id3_1;
        id3v2   *id3_2;

        // First, we'll try to get id3v1 info
        id3_1 = mpeg_audio->GetId3v1Info();

        if (id3_1)
        {
            // We now always use the v2 struct
            m_id3.reset(new struct id3v2);

            // Copy the relevant v1 fields over to the v2 struct
            m_id3->title = id3_1->title;
            m_id3->artist = id3_1->artist;
            m_id3->album = id3_1->album;
            m_id3->year = id3_1->year;
            m_id3->comment = id3_1->comment;

            if (id3_1->track)
            {
                char buf[20];
                sprintf(buf, "%d", id3_1->track);

                m_id3->track = buf;
            }

            if (id3_1->genre_string)
                m_id3->genre = id3_1->genre_string;
        }

        // Try to get id3v2 info
        id3_2 = mpeg_audio->GetId3v2Info();

        if (id3_2 && id3_2->is_id3v2_present)
        {
            if (m_id3.get() == NULL)
            {
                m_id3.reset(new id3v2);
                *m_id3 = *id3_2;
            }
            else
            {
                // Ok, both id3 v1 and v2 info exists in this file.  Our target
                // struct contains the elements of the v1 fields, but we're going
                // to copy over it with any v2 fields which are present.  This
                // will give us the effect of v2 taking precedence, while allowing
                // the v1 data to fill in anything missing from v2.

                if (id3_2->title.length())        m_id3->title         = id3_2->title       ;
                if (id3_2->subtitle.length())     m_id3->subtitle      = id3_2->subtitle    ;
                if (id3_2->artist.length())       m_id3->artist        = id3_2->artist      ;
                if (id3_2->album.length())        m_id3->album         = id3_2->album       ;
                if (id3_2->year.length())         m_id3->year          = id3_2->year        ;
                if (id3_2->comment.length())      m_id3->comment       = id3_2->comment     ;
                if (id3_2->track.length())        m_id3->track         = id3_2->track       ;
                if (id3_2->genre.length())        m_id3->genre         = id3_2->genre       ;
                if (id3_2->record_label.length()) m_id3->record_label  = id3_2->record_label;
                if (id3_2->copyright.length())    m_id3->copyright     = id3_2->copyright   ;
                if (id3_2->date.length())         m_id3->date          = id3_2->date        ;
                if (id3_2->time.length())         m_id3->time          = id3_2->time        ;
            }

        }

        if (m_id3.get() == NULL)
        {
            // Oops, there's no id3 info at all, so we'll create an empty struct
            // since we still need to attempt to read from this pointer later on.

            m_id3.reset(new struct id3v2);
        }
    }
}
////////////////////////////////////////////////////////////////////////////////
id3v2 *AudioDetails::InitId3(id3v2 *&id3, bool full)
{
    try
    {
        if (!id3)
        {
            if (!m_id3.get())
                ScanFileImage(full, false);

            id3 = m_id3.get();
        }
    }
    catch (...)
    {
        throw;  // JJP: Breakpoint test
    }

    return id3;
}
////////////////////////////////////////////////////////////////////////////////
Mp3Info *AudioDetails::InitMp3(Mp3Info *&mp3, bool full)
{
    if (!mp3)
    {
        if (!m_mp3.get())
            ScanFileImage(full, true);

        mp3 = m_mp3.get();
    }

    return mp3;
}
////////////////////////////////////////////////////////////////////////////////
std::string AudioDetails::CalculateDuration(bool accurate)
{
    std::string str;
    Mp3Info *mp3 = NULL;

    // Only if we've already gotten the quick info and we're being asked for
    // the accurate info, do we want to force a re-scan
    if (m_mp3.get() && (!m_accuracy) && (accurate))
    {
        // Reset the autoptr so that the following InitMp3 line will
        // re-scan the file
        m_mp3.reset(NULL);
    }

    if (InitMp3(mp3, accurate))
    {
        char buf[20];
        sprintf(buf,"%u", mp3->duration_ms);
        str = buf;
    }

    return str;
}
////////////////////////////////////////////////////////////////////////////////
std::string AudioDetails::FindDuration(KeyDataType type)
{
    bool accurate = (type == kdtAccurateDuration);
    std::string duration;
    bool update = false;

    std::string cache_token;

#if 0
/* We're unable to use the m_duration member as a shortcut to prevent a file
   cache lookup.  This is because there can be more than one instance of this
   DetailsApi opened on the same url.  e.g. an instance that calculated the quick
   duration can be open as well as an instance that is used when the song is
   played and thus calculated the accurate duration.  Each of these would think
   that they had the correct duration in their copies of m_duration.  So we remove
   the usage of m_duration and instead all instances hit the file cache to see
   if there is a known duration for this url.
*/
    if (m_duration.length())
    {
        // There is already a duration known for this file.

        // If we already have the accurate one known, or the non accurate one
        // is being requested, then use the one we have, else continue
        if ((m_accuracy == true) || (accurate == false))
        {
            return m_duration;
        }
    }
#endif

    // Construct the token used to look up this entry in the disk cache
    cache_token = m_url_api->Url_GetText();
    cache_token += "?TiVoAudioAccuracyAndDuration=1";

    // Local block for the DataRef
    {
        // Attempt to get this token from the cache
        DataRef data = m_cache_api->Cache_GetDataApi(cache_token.c_str(), NULL);

        // Validate the DataApi reference and its size
        if (((bool)data) && data->Data_GetSize())
        {
            // The entry was found in the cache, so we'll use it.

            // Get the data's length
            DWORD len = data->Data_GetSize();

            // Get the data's address
            const char *addr = (const char *)data->Data_Get();

            // Make a local copy of the data
            std::string str_buf(addr, len);

            // Pointer to the local copy
            const char *p = str_buf.c_str();

            // Validate the length, the address, and some data
            if ((len >= 3) && p && p[1] == ':')
            {
                // Check the type of accuracy in the cache entry
                if (p[0] == 'E' && (accurate == true))
                {
                    // We've been asked for the accurate duration, but the cache
                    // entry only contains the estimated duration, so we'll set
                    // the flag indicating that we must re-calculate
                    update = true;
                }
                else
                {
                    // Check the duration value for known bad values
                    if (strtoul(&p[2], NULL, 0) == 0)
                    {
                        update = true;
                    }
                    else
                    {
                        // Ok, at this point, we know that the duration string in
                        // the cache entry is valid and useful.

                        // Get the duration string from the cache entry
                        duration = &p[2];
                    }
                }
            }
            else
            {
                // The length or format of our data is invalid, so cause it to be
                // regenerated.
                update = true;
            }
        }
        else
        {
            // The entry was *NOT* found in the cache so set the update flag
            update = true;
        }
    }

    if (update)
    {
        char buf[25];

        // Calculate the duration
        duration = CalculateDuration(accurate);

        if (duration.length() > 0)
        {

            // Format a string to go into the cache data.  This will contain either
            // "A:xxxx" for accurate duration values, or "E:xxxx" for estimated
            // duration values.
            sprintf(buf, "%c:%s", (accurate ? 'A' : 'E'), duration.c_str());

            // Write it back to the cache
            ApiRef<StringDataApi> str_api = new StringDataApi(buf);

            m_cache_api->Cache_UpdateData(cache_token.c_str(), str_api, NULL);
        }
        else
        {
            // Oops, something bad happened, so we'll return an empty string.
            return "";
        }
    }

    m_duration = duration;

    return duration;
}
////////////////////////////////////////////////////////////////////////////////
LPCSTR  TIVO_API_METHOD AudioDetails::Details_QueryValue(LPCSTR key)
{
    TIVO_API_TRY

    KeyDataType type;

    Mp3Info *mp3 = NULL;
    id3v2   *id3 = NULL;

    type = GetTypeFromString(key);

    // Clear the return string
    m_value = "";

    switch(type)
    {
        case kdtUnknown:
            SetLastError(ERROR_NOT_FOUND);
            return NULL;

        case kdtType:
            m_value = "audio/*";
            break;

        case kdtTitle:
            if (InitId3(id3))
            {
                m_value = Trim(id3->title);

                if (!m_value.length())
                    m_value = GetFileNameFromUrl();
            }
            else
                m_value = GetFileNameFromUrl();
            break;

        case kdtSongTitle:
            if (InitId3(id3))
                m_value = Trim(id3->title);

            break;

        case kdtSourceLocation:
            m_value = SafeString(m_url_api->Url_GetDocument());
            std::replace(m_value.begin(), m_value.end(), '/', '\\');
            break;

        case kdtFileName:
            {
                char name[MAXFILE];
                char ext[MAXEXT];

                // JJP: Hack for now to allow internet radio
                if (strcmp(SafeString(m_url_api->Url_GetScheme()).c_str(), "http") == 0)
                {
                    // Use the entire url for the filename
                    m_value = SafeString(m_url_api->Url_GetText());
                }
                else
                {
                    m_value = ExtractFileName(SafeString(m_url_api->Url_GetDocument()));
                }
            }
            break;


        case kdtSourceFormat:
            m_value = "audio/mpeg";
            break;

        case kdtCreationDate:
        case kdtLastChangeDate:
        case kdtLastAccessDate:
            {
                FILETIME ct, mt, at;
                FILETIME ft;
                int fd;

                // Attempt to open the file again
                if ((fd = open(SafeString(m_url_api->Url_GetDocument()).c_str(), O_RDONLY)) == -1)
                    break;

                // Attempt to get the current times
                if (!GetFileTime((void*)_get_osfhandle(fd), &ct, &at, &mt))
                {
                    close(fd);
                    break;
                }
                close(fd);

                switch(type)
                {
                    case kdtCreationDate:
                        ft = ct;
                        break;
                    case kdtLastChangeDate:
                        ft = mt;
                        break;
                    case kdtLastAccessDate:
                        ft = at;
                        break;
                }

                // If this time component isn't supported, it will come back
                // with all zero values.  In that case skip it.
                if (!ft.dwLowDateTime && !ft.dwHighDateTime)
                    break;

                // Format the time into a string
                const char *p = m_dhelp_api->Details_FormatDate(&ft);
                if (p)
                {
                    m_value = p;
                    GlobalFree((void*)p);
                }
            }
            break;

        case kdtSourceSize:
            if (InitMp3(mp3))
            {
                char buf[20];

                sprintf(buf,"%u",mp3->file_length);
                m_value = buf;
            }
            break;

        case kdtKeywords:
            // Not supported in Id3v1.1 or Id3v2.x
            break;

        case kdtCopyright:
            if (InitMp3(mp3))
            {
                m_value = mp3->copyright ? "Copyright" : "";
            }
            break;

        case kdtSourceBitRate:
            if (InitMp3(mp3))
            {
                char buf[20];

                sprintf(buf,"%d",mp3->bitrate_kbps * 1000);
                m_value = buf;
            }
            break;

        case kdtSourceSampleRate:
            if (InitMp3(mp3))
            {
                char buf[20];

                sprintf(buf,"%d",mp3->sample_hz);
                m_value = buf;
            }
            break;

        case kdtMusicGenre:
            if (InitId3(id3))
                m_value = Trim(id3->genre);
            break;

        case kdtArtistName:
            if (InitId3(id3))
                m_value = Trim(id3->artist);
            break;

        case kdtAlbumTitle:
            if (InitId3(id3))
                m_value = Trim(id3->album);
            break;

        case kdtDuration:
        case kdtAccurateDuration:
            m_value = FindDuration(type);
            break;

        case kdtAlbumYear:
            if (InitId3(id3))
                m_value = Trim(id3->year);
            break;

        case kdtRecordLabel:
            if (InitId3(id3))
                m_value = Trim(id3->record_label);
            break;
    }

    if (!m_value.length())
    {
        SetLastError(ERROR_SUCCESS);
        return NULL;
    }

    return m_value.c_str();

    TIVO_API_CATCH_RETURN(NULL)
}
////////////////////////////////////////////////////////////////////////////////
KeyDataType AudioDetails::GetTypeFromString(LPCSTR str)
{
    int j;

    for(j=0;j<(sizeof(s_key_array) / sizeof(s_key_array[0]));j++)
    {
        if (strcmp(str,s_key_array[j].key) == 0)
            return s_key_array[j].type;
    }

    return kdtUnknown;
}
////////////////////////////////////////////////////////////////////////////////
std::string AudioDetails::GetFileNameFromUrl()
{
    char name[MAXFILE], *p;

    // JJP: Hack for now to allow internet radio
    if (strcmp(SafeString(m_url_api->Url_GetScheme()).c_str(), "http") == 0)
    {
        return SafeString(m_url_api->Url_GetText());
    }

    strcpy(name, ExtractFileNameOnly(SafeString(m_url_api->Url_GetDocument())).c_str());

    // Convert any underbars to spaces
    do
    {
        p = strchr(name, '_');
        if (p)
            *p = ' ';

    }
    while(p);

    return std::string(name);
}
////////////////////////////////////////////////////////////////////////////////
