/******************************************************************************

    TiVo Desktop Mp3.dll Source Code and Build Process
    Copyright (c) 2003 TiVo Inc.

    This library is free software; you can redistribute it and/or modify it
    under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation; either version 2.1 of the License, or (at
    your option) any later version.

    This library is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser
    General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this library; if not, write to the Free Software Foundation,
    Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307 USA.

    April 4th, 2003

    www.tivo.com

    TiVo Inc.
    P.O. Box 2160
    Alviso, CA 95002

******************************************************************************/

#ifndef INCLUDE_GUARD_9CCC4867_0EBC_11D5_8018_00A0CCD13742
#define INCLUDE_GUARD_9CCC4867_0EBC_11D5_8018_00A0CCD13742

//////////////////////////////////////////////////////////////////////////////////////////////////
//
// ApiHelp.h
//
// SYNOPSIS
//
//  Utility templates for implementing the common code required by every Api subtype (creation,
//  reference couting, etc.).
//
//
// EXAMPLE
//
//  Quick implementation of standard reference counting for FooApi is accomplished as follows.
//
//      using namespace TivoApi;
//
//      class MyFoo: public RefCounted<FooApi>
//      {
//          // Your implementation of FooApi goes here.
//      };
//          .
//          .
//
//      FooApi* foo = new MyFoo();              // Reference count automatically 1.
//
//      // Use FooApi here.
//
//      foo->Api_Release();                     // MyFoo instance deleted.
//
//
//  Implementation of a reference-counted singleton is accomplished as follows.  The instance
//  of BarApi is retrieved via the static function MyBar::GetApi(), which is implemented for
//  you by the Singleton template.
//
//      class MyBar: public RefCounted< Singleton<MyBar, BarApi> > { ... };
//
//          .
//          .
//
//      BarApi* bar1 = MyBar::GetApi();         // Creates "the" instance of MyBar with
//                                              // reference count 1.
//      // Use BarApi here.
//
//      BarApi* bar2 = MyBar::GetApi();         // Returns the same instance, but reference
//                                              // count now 2.
//
//      // Use "second" instance of BarApi here.
//
//      bar1->Api_Release();
//      bar2->Api_Release();                    // MyBar instance finally destroyed.
//
//
// HISTORY
//
//  marnold, 3/8/2001 - Initial version.
//  marnold, 3/9/2001 - Revised comments.
//  marnold, 5/1/2001 - Changed namespace to TivoApi.
//
//
// Copyright (c) 2001 TiVo Inc., all rights reserved.
//
//////////////////////////////////////////////////////////////////////////////////////////////////


#include "TivoApi/Api.h"

#include "Windows.h"


namespace TivoApi {


//////////////////////////////////////////////////////////////////////////////////////////////////
//
// RefCounted
//
// Implements Api with a standard reference counting scheme.  Use for most classes of Api objects
// that are created and destroyed on demand.
//
//////////////////////////////////////////////////////////////////////////////////////////////////

template <class T>
class RefCounted:

	public T

{
	public:

		virtual void TIVO_API_METHOD Api_AddRef()
		{
			InterlockedIncrement(&ref_count);
		}

		virtual void TIVO_API_METHOD Api_Release()
		{
			if (InterlockedDecrement(&ref_count) > 0)
                return;

            delete this;
		}


	protected:

		RefCounted(): ref_count(1) { }

		virtual ~RefCounted() { }


	private:

		LONG ref_count;

};


//////////////////////////////////////////////////////////////////////////////////////////////////
//
// RefNeutral
//
// Implements Api with a "no-op" reference counting scheme.  Use for classes of Api objects that
// exist for the lifetime of the program (i.e., are never destroyed).
//
//////////////////////////////////////////////////////////////////////////////////////////////////

template <typename T_Api>
struct RefNeutral: T_Api

{
	virtual void TIVO_API_METHOD Api_AddRef() { }
	virtual void TIVO_API_METHOD Api_Release() { }
};


//////////////////////////////////////////////////////////////////////////////////////////////////
//
// Instanced
//
// Implements a static GetApi() member that creates a new instance of T each time it is called.
//
//////////////////////////////////////////////////////////////////////////////////////////////////

template <typename T, typename T_Api>
class Instanced:

	public T_Api

{
	public:

		static T* GetApi()
		{
			return new T();
		}

};


//////////////////////////////////////////////////////////////////////////////////////////////////
//
// Singleton
//
// Implements a static GetApi() member that creates a new instance of T only if one currently
// does not exist.  Otherwise, the existing instance's reference count is simply incremented.
//
//////////////////////////////////////////////////////////////////////////////////////////////////

template <typename T, typename T_Api>
class Singleton:

	public T_Api

{
	public:

		static T* GetApi()
		{
			if (singleton)
				singleton->Api_AddRef();

			else
				singleton = new T();

			return singleton;
		}


	protected:

		~Singleton()
		{
			singleton = NULL;
		}


	private:

		static T* singleton;

};

template <typename T, typename T_Api>
T* Singleton<T, T_Api>::singleton;


} // TivoApi


#endif // INCLUDE_GUARD
