/******************************************************************************

    TiVo Desktop Mp3.dll Source Code and Build Process
    Copyright (c) 2003 TiVo Inc.

    This library is free software; you can redistribute it and/or modify it
    under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation; either version 2.1 of the License, or (at
    your option) any later version.

    This library is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser
    General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this library; if not, write to the Free Software Foundation,
    Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307 USA.

    April 4th, 2003

    www.tivo.com

    TiVo Inc.
    P.O. Box 2160
    Alviso, CA 95002

******************************************************************************/

//////////////////////////////////////////////////////////////////////////////////////////////////
//
// ApiManager.cpp
//
//
// Copyright (c) 2001 TiVo, Inc., all rights reserved.
//
//////////////////////////////////////////////////////////////////////////////////////////////////


#include "PreCompiled.h"

#include "TivoApi/ApiManager.h"

#include <algorithm>
#include <functional>
#include <stdexcept>
#include <memory>

#include <assert.h>
#include <string.h>
#include <dir.h>

using namespace TivoApi;
using namespace std;


namespace {


//////////////////////////////////////////////////////////////////////////////////////////////////

enum { c_MaxTypeIdSize = 100 };


//////////////////////////////////////////////////////////////////////////////////////////////////
//
// Given a file path string in the form "C:\\Program Files\\Notepad.exe", returns just the
// folder portion, as in "C:\\Program Files\\".
//
//////////////////////////////////////////////////////////////////////////////////////////////////

void SplitFolder(const char* path, char* folder)
{
    char drive[MAXDRIVE];
    char dir[MAXDIR];

    fnsplit(path, drive, dir, NULL, NULL);

    strcpy(folder, drive);
    strcat(folder, dir);
}


//////////////////////////////////////////////////////////////////////////////////////////////////
//
// Library
//
// Represents a loaded Windows library; a DLL.
//
//////////////////////////////////////////////////////////////////////////////////////////////////

class Library
{
    public:

        // Creators

        Library();
        Library(const char* path);

        ~Library();


        // Accessors

        bool IsValid() const;

        const HMODULE hModule;

};


//////////////////////////////////////////////////////////////////////////////////////////////////

inline Library::Library():
    hModule(NULL)
{
}

Library::Library(const char* path):
    hModule(LoadLibraryA(path))
{
}

Library::~Library()
{
    FreeLibrary(hModule);
}


//////////////////////////////////////////////////////////////////////////////////////////////////

inline bool Library::IsValid() const
{
    return int(hModule) > int(HINSTANCE_ERROR);
}


} // unnamed


//////////////////////////////////////////////////////////////////////////////////////////////////
//
// Module
//
// Each instance of this class represents one loaded Api "module".
//
//////////////////////////////////////////////////////////////////////////////////////////////////

class ApiManager::Module
{
    public:

        // Creators

        private: Module(); public: // ???

        Module(const char* path, DllCallbackApi* callback);

        ~Module();


        // Accessors

        bool                IsValid() const;

        bool                HasRefs() const;

        HMODULE             GetHandle() const;



        // Manipulators

        Api*                GetApi(const char* type_id);

        void                AddRef();
        void                Release();


    private:

        // Internals

        Library             m_lib;
        TivoGetApi          m_GetApi;
        ApiRef<DllApi>      m_api;
        int                 m_api_count;
        bool                m_initialized;

};


//////////////////////////////////////////////////////////////////////////////////////////////////
//
// This ctor form creates a "zombie" Module instance that fails all Api requests.
//
//////////////////////////////////////////////////////////////////////////////////////////////////

ApiManager::Module::Module():
    m_GetApi(NULL),
    m_api_count(0),
    m_initialized(true)
{
}


//////////////////////////////////////////////////////////////////////////////////////////////////
//
// When supplied with the path of a DLL that implements the expected entry-points, etc., this
// ctor form creates a valid Module instance.  After this constructor executes, check IsValid()
// to see if the DLL was loaded, initialized, and is ready to start hosting Api instances via
// the Tivo_GetApi() entry-point.
//
//////////////////////////////////////////////////////////////////////////////////////////////////

ApiManager::Module::Module(const char* path, DllCallbackApi* callback):
    m_lib(path),
    m_GetApi(NULL),
    m_api_count(0),
    m_initialized(false)
{
    if (!m_lib.IsValid())
        return;

    m_GetApi = reinterpret_cast<TivoGetApi>(GetProcAddress(m_lib.hModule, "Tivo_GetApi"));

    m_api = GetApi("DllApi");

    if (!m_api.Get())
        return;

    if (!m_api->Dll_Initialize(callback))
        return;

    m_initialized = true;
}


//////////////////////////////////////////////////////////////////////////////////////////////////

ApiManager::Module::~Module()
{
    assert(!HasRefs()); // Oops, our DLL still has "live" Api instances out there?!
}


//////////////////////////////////////////////////////////////////////////////////////////////////

inline bool ApiManager::Module::IsValid() const
{
    return m_initialized;
}

inline bool ApiManager::Module::HasRefs() const
{
    return m_api_count > 0;
}

inline HMODULE ApiManager::Module::GetHandle() const
{
    return m_lib.hModule;
}


//////////////////////////////////////////////////////////////////////////////////////////////////
//
// Returns a pointer to a valid instance of the Api requested.  NULL is returned if the Api is
// not supported.
//
//////////////////////////////////////////////////////////////////////////////////////////////////

inline Api* ApiManager::Module::GetApi(const char* type_id)
{
    return m_GetApi ? m_GetApi(type_id) : NULL;
}


//////////////////////////////////////////////////////////////////////////////////////////////////
//
// Used to notify a Module instance of the comings and goings of its associated Api instances.
//
//////////////////////////////////////////////////////////////////////////////////////////////////

inline void ApiManager::Module::AddRef()
{
    m_api_count++;
}

inline void ApiManager::Module::Release()
{
    if (!m_api_count)
        return;

    m_api_count--;
}


namespace {


//////////////////////////////////////////////////////////////////////////////////////////////////
//
// Some Module-related utility functions referenced later in this file.
//
//////////////////////////////////////////////////////////////////////////////////////////////////

bool UnloadModule(ApiManager::Module* module)
{
    if (module->HasRefs())
        return false;

    delete module;

    return true;
}
bool IsMatchingModule(const ApiManager::Module* module, HMODULE hModule)
{
    return module->GetHandle() == hModule;
}


} // unnmamed


//////////////////////////////////////////////////////////////////////////////////////////////////
//
// ApiManager
//
//////////////////////////////////////////////////////////////////////////////////////////////////

ApiManager::ApiManager() { }


//////////////////////////////////////////////////////////////////////////////////////////////////

ApiManager::~ApiManager()
{
    UnloadDlls();

    assert(m_modules.empty()); // Oops, why wasn't everyone unloaded?!
}


//////////////////////////////////////////////////////////////////////////////////////////////////
//
// Returns a string containing the full name of the DLL that houses the implementation of the
// given Api.  The string will remain valid until the next call to GetApiModulePath().
//
//////////////////////////////////////////////////////////////////////////////////////////////////

const char* ApiManager::GetApiModulePath(ApiRef<Api> api) const
{
    Module* module = GetModuleByApi(api.Get());

    strcpy(m_temp, "");

    if (module)
        GetModuleFileNameA(module->GetHandle(), m_temp, MAX_PATH);

    return m_temp;
}


//////////////////////////////////////////////////////////////////////////////////////////////////
//
// Returns a pointer to a valid instance of the Api requested.  NULL is returned if the DLL
// could not be loaded or if the Api is not supported.
//
//////////////////////////////////////////////////////////////////////////////////////////////////

ApiRef<Api> ApiManager::GetApi(const char* type_id, const char* path)
{
    assert(!IsBadStringPtrA(type_id, c_MaxTypeIdSize));
    assert(!IsBadStringPtrA(path, MAX_PATH));

    Module* module = GetModuleByPath(path);

    if (!module)
        return NULL;

    ApiRef<Api> new_api = module->GetApi(type_id);

    if (!new_api.Get())
        return NULL;

    if (!GetModuleByApi(new_api.Get()))
        module->AddRef();

    m_api_map[new_api.Get()] = module;

    return new_api;
}


//////////////////////////////////////////////////////////////////////////////////////////////////
//
// Returns one or more pointers to valid instances of the Api requested.  An empty array is
// returned if no DLL could be loaded or if the Api is not supported.
//
//////////////////////////////////////////////////////////////////////////////////////////////////

ApiManager::ApiArray ApiManager::SearchApi(const char* type_id, const char* path)
{
    assert(!IsBadStringPtrA(type_id, c_MaxTypeIdSize));
    assert(!IsBadStringPtrA(path, MAX_PATH));

    HANDLE hFind = NULL;

    try
    {
        ApiArray new_apis;

        char folder[MAX_PATH];

        SplitFolder(path, folder);

        WIN32_FIND_DATAA data;

        hFind = FindFirstFileA(path, &data);

        for (bool iter = (hFind != INVALID_HANDLE_VALUE); iter; iter = FindNextFileA(hFind, &data))
        {
            char match[MAX_PATH];

            strcpy(match, folder);
            strcat(match, data.cFileName);

            ApiRef<Api> new_api = GetApi(type_id, match);

            if (!new_api.Get())
                continue;

            new_apis.push_back(new_api);
        }

        FindClose(hFind);

        return new_apis;
    }

    catch (...)
    {
        FindClose(hFind);

        return ApiArray();
    }
}


//////////////////////////////////////////////////////////////////////////////////////////////////
//
// Removes from memory all DLLs that currently have no associated outstanding Api instances.
//
//////////////////////////////////////////////////////////////////////////////////////////////////

void ApiManager::UnloadDlls()
{
    ModuleArray::iterator begin = m_modules.begin();
    ModuleArray::iterator end = m_modules.end();

    m_modules.erase(remove_if(begin, end, UnloadModule), end);
}


//////////////////////////////////////////////////////////////////////////////////////////////////
//
// Called by each DLL loaded when the final Api_Release() occurs for one of its associated Api
// instances.  Note that it is legal for DLLs to call this function for Api instances that were
// never returned by their Tivo_GetApi() entry-point.  We simply ignore the call in such cases.
//
// WARNING!  We must not let any C++ exceptions leak back to the calling DLL (whose exception-
// handling scheme may not necessarily be compatible with ours).
//
//////////////////////////////////////////////////////////////////////////////////////////////////

void TIVO_API_METHOD ApiManager::Dll_OnApiRelease(Api* old_api)
{
    assert(!IsBadWritePtr(old_api, sizeof(Api)));

    try
    {
        Module* module = GetModuleByApi(old_api);

        if (module)
            module->Release();

        m_api_map.erase(old_api);
    }

    catch (...) { }
}


//////////////////////////////////////////////////////////////////////////////////////////////////
//
// Given a valid path to a DLL that implements the expected entry-points, etc., returns the
// Module instance that represents it.  If the DLL is not currently loaded, this function will
// load it.  Returns NULL otherwise.
//
//////////////////////////////////////////////////////////////////////////////////////////////////

ApiManager::Module* ApiManager::GetModuleByPath(const char* path)
{
    assert(!IsBadStringPtrA(path, MAX_PATH));

    Library lib(path);

    if (!lib.IsValid())
        return NULL;

    ModuleArray::iterator match = find_if
    (
        m_modules.begin(),
        m_modules.end(),
        bind2nd(ptr_fun(IsMatchingModule), lib.hModule)
    );

    if (match != m_modules.end())
        return *match;

    auto_ptr<Module> new_mod(new Module(path, this));

    if (!new_mod->IsValid())
        return NULL;

    m_modules.push_back(new_mod.get());

    return new_mod.release();
}


//////////////////////////////////////////////////////////////////////////////////////////////////
//
// Returns the Module instance associated with the given Api instance.  Returns NULL is the Api
// is not associated with any DLL known to this ApiManager.
//
//////////////////////////////////////////////////////////////////////////////////////////////////

ApiManager::Module* ApiManager::GetModuleByApi(Api* api) const
{
    ApiToModuleMap::const_iterator match = m_api_map.find(api);

    if (match == m_api_map.end())
        return NULL;

    return match->second;
}


