/******************************************************************************

    TiVo Desktop Mp3.dll Source Code and Build Process
    Copyright (c) 2003 TiVo Inc.

    This library is free software; you can redistribute it and/or modify it
    under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation; either version 2.1 of the License, or (at
    your option) any later version.

    This library is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser
    General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this library; if not, write to the Free Software Foundation,
    Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307 USA.

    April 4th, 2003

    www.tivo.com

    TiVo Inc.
    P.O. Box 2160
    Alviso, CA 95002

******************************************************************************/

#ifndef INCLUDE_GUARD_D2C91100_0468_11D5_A849_444553540000
#define INCLUDE_GUARD_D2C91100_0468_11D5_A849_444553540000

//////////////////////////////////////////////////////////////////////////////////////////////////
//
// ApiRef.h
//
// SYNOPSIS
//
//  A "smart pointer" helper class that helps automate most of the lifetime management chores
//  required when using subtypes of Api.
//
//
// DETAILS
//
//  For every ApiRef that refers to a valid Api instance, the Api's reference count increments
//  by 1.  When 'ref1', 'ref2', and 'ref3' each all finally go out of scope, the Api instance's
//  reference count will drop to 0, causing the instance to be destroyed.
//
//      ApiRef<SomeApi> ref1(GetSomeApi()); // Reference count starts at 1.
//      ApiRef<SomeApi> ref2(ref2);         // Reference count now 2.
//      ApiRef<SomeApi> ref3;
//      ref3 = ref2                         // Reference count now 3.
//
//
//  When a "raw" Api instance pointer is used to initialize or assign an ApiRef, ownership of
//  the instance is assumed "trasferred" to the ApiRef.  That is, the Api instance's reference
//  count will remain unchanged.
//
//      SomeApi* raw1 = GetSomeApi();       // Reference count starts at 1.
//      ApiRef<SomeApi> ref4 = raw1;        // Reference count stays 1.
//
//      SomeApi* raw2 = GetSomeApi();       // Reference count starts at 1.
//      ApiRef<SomeApi> ref5;
//      ref5 = raw2;                        // Reference count stays 1.
//
//
//  A valid ApiRef can also be automatically converted to a "raw" Api instance pointer, which
//  has the side-effect of incrementing the Api's reference count by 1.  That is, the receiver
//  of the raw pointer resulting from the conversion is assumed to participate in ownership.
//
//      SomeApi* raw3;
//      raw3 = ref1;                        // Reference count now 4 ('raw3' represents a
//                                          // seperate reference to the Api instance).
//
//  A valid ApiRef instance can be dereferenced just like a built-in pointer (a la auto_ptr).
//  It is a programming error to dereference an invaild (uninitialized or unassigned) ApiRef.
//
//      ref1->SomeApiMember();
//
//      ApiRef<SomeApi> bad;
//      bad->SomeApiMember();               // ERROR!
//
//
// HISTORY
//
//  marnold, 2/21/2001 - Initial version.
//  marnold, 3/6/2001 - Added Get() member.
//  marnold, 3/7/2001 - Removed operator&() -- too troublesome.
//
//
// Copyright (c) 2001 TiVo, Inc., all rights reserved.
//
//////////////////////////////////////////////////////////////////////////////////////////////////


#include "TivoApi/Api.h"

#include "Windows.h"

#include <assert.h>
#include <stddef.h>


namespace TivoApi {


//////////////////////////////////////////////////////////////////////////////////////////////////

template <typename T>
class ApiRef
{
    public:

        // Creators

        ApiRef(Api* new_api = NULL);
        ApiRef(const ApiRef& other);

        ~ApiRef();


        // Accessors

        T*              Get() const;

                        operator bool() const;

#ifdef __BORLANDC__
        bool            operator!() const;
#endif

                        operator T*() const;
                        operator ApiRef<Api>() const;

        T*              operator->() const;

        bool            operator==(const ApiRef& rhs) const;
        bool            operator!=(const ApiRef& rhs) const;


        // Manipulators

/* ???
        const ApiRef&   operator=(Api* rhs);
*/
        const ApiRef&   operator=(const ApiRef& rhs);


    private:

        // Internals

        void            AddRef() const;
        void            Release() const;

        T*              api;

};


//////////////////////////////////////////////////////////////////////////////////////////////////

template <typename T>
inline void ApiRef<T>::AddRef() const
{
    if (api)
        api->Api_AddRef();
}

template <typename T>
inline void ApiRef<T>::Release() const
{
    if (api)
        api->Api_Release();
}


//////////////////////////////////////////////////////////////////////////////////////////////////

template <typename T>
inline ApiRef<T>::ApiRef(Api* new_api):
    api(static_cast<T*>(new_api))
{
}

template <typename T>
inline ApiRef<T>::ApiRef(const ApiRef& other):
    api(other.api)
{
    AddRef();
}


//////////////////////////////////////////////////////////////////////////////////////////////////

template <typename T>
inline ApiRef<T>::~ApiRef()
{
    Release();
}


//////////////////////////////////////////////////////////////////////////////////////////////////

template <typename T>
inline T* ApiRef<T>::Get() const
{
    return api;
}


//////////////////////////////////////////////////////////////////////////////////////////////////

template <typename T>
inline ApiRef<T>::operator bool() const
{
    return api;
}


#ifdef __BORLANDC__

//////////////////////////////////////////////////////////////////////////////////////////////////
//
// This is a workaround for a bug in Borland compilers.  For some reason, the compiler is unable
// to properly select the "ApiRef<T>::operator bool()" conversion for negated if-statements.
//
//      ApiRef<Foo> x;
//
//      if (x)      // OK under Borland, "operator bool()" called.
//          ...
//
//      if (!x)     // Ambiguous under Borland if "operator!()" is not also defined for 'x'.
//          ...
//
// Other compilers, such as Comeau Computing's, invoke "operator bool()" for both if-statements.
//
//////////////////////////////////////////////////////////////////////////////////////////////////

template <typename T>
inline bool ApiRef<T>::operator!() const
{
    return !operator bool();
}

#endif


//////////////////////////////////////////////////////////////////////////////////////////////////

template <typename T>
inline ApiRef<T>::operator T*() const
{
    AddRef();

    return api;
}

template <typename T>
inline ApiRef<T>::operator ApiRef<Api>() const
{
    AddRef();

    return api;
}


//////////////////////////////////////////////////////////////////////////////////////////////////

template <typename T>
inline T* ApiRef<T>::operator->() const
{

/*
#ifdef _DEBUG

    DWORD temp = GetLastError();

    assert(!IsBadWritePtr(api, sizeof(Api)));

    SetLastError(temp);

#endif
*/

    return api;
}


//////////////////////////////////////////////////////////////////////////////////////////////////

template <typename T>
inline bool ApiRef<T>::operator==(const ApiRef& rhs) const
{
    return api == rhs.api;
}

template <typename T>
inline bool ApiRef<T>::operator!=(const ApiRef& rhs) const
{
    return api != rhs.api;
}


//////////////////////////////////////////////////////////////////////////////////////////////////

/* ???
template <typename T>
inline const ApiRef<T>& ApiRef<T>::operator=(Api* rhs)
{
    Release();

    api = static_cast<T*>(rhs);

    return *this;
}
*/

template <typename T>
const ApiRef<T>& ApiRef<T>::operator=(const ApiRef& rhs)
{
    if (this != &rhs)
    {
        Release();

        api = rhs.api;

        AddRef();
    }

    return *this;
}


} // TivoApi


#endif // INCLUDE_GUARD