/******************************************************************************

    TiVo Desktop Mp3.dll Source Code and Build Process
    Copyright (c) 2003 TiVo Inc.

    This library is free software; you can redistribute it and/or modify it
    under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation; either version 2.1 of the License, or (at
    your option) any later version.

    This library is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser
    General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this library; if not, write to the Free Software Foundation,
    Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307 USA.

    April 4th, 2003

    www.tivo.com

    TiVo Inc.
    P.O. Box 2160
    Alviso, CA 95002

******************************************************************************/

///////////////////////////////////////////////////////////////////////////////
//
// id3v2.cpp
//
// This provides routines for id3v2 tags using Id3Lib
//
// Copyright (c) 2002 TiVo Inc., all rights reserved.
//
///////////////////////////////////////////////////////////////////////////////

#include <stdio.h>
#include <string.h>
#include <io.h>
#include <memory>
#include <ctype.h>

#include "id3v2.h"
#include "id3v1.h" // Needed for genre string converter

#pragma option push -w-8071 -w-8027
#include "Id3/Tag.h"
#include "Id3/Misc_Support.h"
#include "Id3/Readers.h"
#pragma option pop

///////////////////////////////////////////////////////////////////////////////
bool ProcessTag(ID3_Tag &Tag, struct id3v2 *id3)
{
    try
    {
        id3->is_id3v2_present = true;

        std::auto_ptr<ID3_Tag::ConstIterator> iter((ID3_Tag::ConstIterator*)Tag.CreateIterator());

        const ID3_Frame *frame;

        while ((frame = iter->GetNext()) != NULL)
        {
            char *cp_text = ID3_GetString(frame, ID3FN_TEXT);

            std::string text;

            if (cp_text)
                text = cp_text;

            delete [] cp_text;

            ID3_FrameID frame_id = frame->GetID();

#ifdef DUMP_ALL_TAGS

            static bool first = true;

            if (first)
            {
                first = false;
                printf("\nDump of all tags:\n-----------------\n");
            }
#if 0
            const char* desc = frame->GetDescription();

            if (!desc)
                desc = "";

            printf("%s(%3d)     :%s: \"%s\"\n", frame->GetTextID(), frame_id, desc, text.c_str());
#else
            printf("%s(%3d)     :%s\n", frame->GetTextID(), frame_id, text.c_str());
#endif // 0
#endif // DUMP_ALL_TAGS

            switch (frame_id)
            {
                case ID3FID_TITLE:
                    id3->title = text;
                    break;

                case ID3FID_SUBTITLE:
                    id3->subtitle = text;
                    break;

                case ID3FID_COMPOSER:
                    // If there's no artist set, then use composer
                    if (id3->artist.length() == 0)
                        id3->artist = text;
                    break;

                case ID3FID_LEADARTIST:
                    // Always use lead artist if is't present
                    id3->artist = text;
                    break;

                case ID3FID_ALBUM:
                    id3->album = text;
                    break;

                case ID3FID_YEAR:
                    id3->year = text;
                    break;

                case ID3FID_COMMENT:
                    id3->comment = text;
                    break;

                case ID3FID_TRACKNUM:
                    id3->track = text;
                    break;

                case ID3FID_CONTENTTYPE:

                    // Check for the numeric representation of id3v1 genre index
                    if ((text[0] == '(') && isdigit(text[1]))
                    {
                        // The format is "(ddd)" where ddd are digits
                        int index = atoi(&text.c_str()[1]);
                        char buf[30];
                        const char *p = "";

                        if (index < 256)
                        {
                            p = GetId3GenreStringFromIndex(index);

                            if (p == NULL)
                            {
                                sprintf(buf, "Unknown (%d)", index);
                                p = buf;
                            }
                        }
                        id3->genre = p;
                    }
                    else
                        id3->genre = text;

                    break;

                case ID3FID_PUBLISHER:
                    id3->record_label = text;
                    break;

                case ID3FID_COPYRIGHT:
                    id3->copyright = text;
                    break;

                case ID3FID_DATE:
                    id3->date = text;
                    break;

                case ID3FID_TIME:
                    id3->time = text;
                    break;
            }
        }
    }
    catch(...)
    {
        return false;
    }

    return true;
}
///////////////////////////////////////////////////////////////////////////////
#pragma argsused
bool GetId3v2FromMp3Image(long flength, const void *vptr, struct id3v2 *id3)
{
    // Validate the struct
    if (!id3)
        return false;

    // Validate the image pointer and its length
    if (vptr == NULL || flength < ID3_TAGHEADERSIZE)
        return false;

    unsigned char *data   = (unsigned char *)vptr;
    bool           ret;

    try
    {
        // Create an ID3 memory reader and attach it to our image
        ID3_MemoryReader mr(data, flength);
        ID3_Tag          tag;

        // Attempt to find a tag
        if (tag.Parse(mr))
        {
            // Found one
            ret = ProcessTag(tag, id3);

            // See if we parsed anything
            if (!ret)
            {
                id3->is_id3v2_present = false;
                return true;
            }
        }
        else
        {
            id3->is_id3v2_present = false;
            return true;
        }

    }
    catch (...)
    {
        ret = false;
    }

    return ret;
}
///////////////////////////////////////////////////////////////////////////////
bool GetId3v2FromMp3File(const char *filename, struct id3v2 *id3)
{
    // Validate the struct
    if (!id3)
        return false;

    // Validate the file
    if (access(filename, 0) != 0)
        return false;

    ID3_Tag tag;
    bool    ret;

    try
    {
        tag.Link(filename, ID3TT_ALL);

        ret = ProcessTag(tag, id3);

        // See if we parsed anything
        if (!ret)
        {
            id3->is_id3v2_present = false;
            return true;
        }
    }
    catch(...)
    {
        ret = false;
    }

    return ret;
}
///////////////////////////////////////////////////////////////////////////////
#ifdef USE_MAIN

#include <errno.h>

int main(int argc, char *argv[])
{
    struct id3v2 id3;

    printf("ID3v2 dumper version 1.00 Copyright (c) 2002 TiVo Inc.\n\n");

    if (argc < 2)
    {
        printf("Usage: id3v2.exe mp3filename\n");
        return 0;
    }

    printf("file:\n-----\n%s\n", argv[1]);

#ifdef USE_FILE_IMAGE
    FILE *fp = fopen(argv[1], "rb");

    if (!fp)
    {
        perror(argv[1]);
        return 1;
    }

    unsigned long len = filelength(fileno(fp));

    char *filebuf = new char[len+1];

    fread(filebuf, 1, len, fp);
    fclose(fp);

    if (GetId3v2FromMp3Image(len, filebuf, &id3) == false)
    {
        printf("Error getting id3 version 2 info %s\n", sys_errlist[errno]);
        return 1;
    }

#else
    if (GetId3v2FromMp3File(argv[1], &id3) == false)
    {
        printf("Error getting id3 version 2 info %s\n", sys_errlist[errno]);
        return 1;
    }
#endif

    if (!id3.is_id3v2_present)
    {
        printf("No id3 version 2 info was found\n");
        return 2;
    }

    printf("\nCollected information:\n----------------------\n");

    if (id3.title.length())        printf("title         :%s\n", id3.title.c_str());
    if (id3.subtitle.length())     printf("subtitle      :%s\n", id3.subtitle.c_str());
    if (id3.artist.length())       printf("artist        :%s\n", id3.artist.c_str());
    if (id3.album.length())        printf("album         :%s\n", id3.album.c_str());
    if (id3.year.length())         printf("year          :%s\n", id3.year.c_str());
    if (id3.track.length())        printf("track         :%s\n", id3.track.c_str());
    if (id3.genre.length())        printf("genre         :%s\n", id3.genre.c_str());
    if (id3.record_label.length()) printf("record_label  :%s\n", id3.record_label.c_str());
    if (id3.copyright.length())    printf("copyright     :%s\n", id3.copyright.c_str());
    if (id3.date.length())         printf("date          :%s\n", id3.date.c_str());
    if (id3.time.length())         printf("time          :%s\n", id3.time.c_str());
    if (id3.comment.length())      printf("comment       :%s\n", id3.comment.c_str());

    return 0;
}
#endif
///////////////////////////////////////////////////////////////////////////////
