/******************************************************************************

    TiVo Desktop Mp3.dll Source Code and Build Process
    Copyright (c) 2003 TiVo Inc.

    This library is free software; you can redistribute it and/or modify it
    under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation; either version 2.1 of the License, or (at
    your option) any later version.

    This library is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser
    General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this library; if not, write to the Free Software Foundation,
    Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307 USA.

    April 4th, 2003

    www.tivo.com

    TiVo Inc.
    P.O. Box 2160
    Alviso, CA 95002

******************************************************************************/

///////////////////////////////////////////////////////////////////////////////
//
// mp3hdr.cpp
//
// This provides routines for accessing the info in mp3 frame headers
//
// Copyright (c) 2002 TiVo Inc., all rights reserved.
//
///////////////////////////////////////////////////////////////////////////////

#include <stdio.h>
#include <string.h>
#include <assert.h>

#ifdef __BORLANDC__
#include <io.h>
#else
#include <sys/stat.h>
#endif

#include "mp3hdr.h"

///////////////////////////////////////////////////////////////////////////////
static long GetFileLengthFromDescriptor(int fd);
///////////////////////////////////////////////////////////////////////////////
static int MpegBitRateArray /*version*/ [3] /*layer*/ [3] /*index*/[16] =
{
    // Version 1
    {
        // Layer 1
        {0,32,64,96,128,160,192,224,256,288,320,352,384,416,448,0},

        // Layer 2
        {0,32,48,56,64,80,96,112,128,160,192,224,256,320,384,0},

        // Layer 3
        {0,32,40,48,56,64,80,96,112,128,160,192,224,256,320,0},

    },

    // Version 2
    {
        // Layer 1
        {0,32,48,56,64,80,96,112,128,144,160,176,192,224,256,0},

        // Layer 2
        {0,8,16,24,32,40,48,56,64,80,96,112,128,144,160,0},

        // Layer 3
        {0,8,16,24,32,40,48,56,64,80,96,112,128,144,160,0},
    },

    // Version 2.5
    {
        // Layer 1
        {0,32,48,56,64,80,96,112,128,144,160,176,192,224,256,0},

        // Layer 2
        {0,8,16,24,32,40,48,56,64,80,96,112,128,144,160,0},

        // Layer 3
        {0,8,16,24,32,40,48,56,64,80,96,112,128,144,160,0},
    },

};

static int MpegLayerArray [4] =
{
    0,
    3,
    2,
    1
};

static int MpegSampleArray /*version*/[3] /*index*/ [4] =
{
    // Version 1
    {44100, 48000, 32000, 0},

    // Version 2
    {22050, 24000, 16000, 0},

    // Version 2.5
    {11025, 12000, 8000, 0},
};
///////////////////////////////////////////////////////////////////////////////
static long GetFileLengthFromDescriptor(int fd)
{
#ifdef __BORLANDC__
    return filelength(fd);
#else
    struct stat s;

    fstat(fd, &s);

    return (long)s.st_size;
#endif
}
///////////////////////////////////////////////////////////////////////////////
static bool FalseWithError(const char *err_str)
{
#ifdef PRINT_ERRORS
    printf("Error: %s\n", err_str);
#else
    err_str;  // To remove the not used warning
#endif
    return false;
}
///////////////////////////////////////////////////////////////////////////////
static bool ReadMp3FrameHeader(char *buf, Mp3Info *mp3)
{
    int j;

    union
    {
        unsigned long   l;
        bits08          raw[4];
        Mp3FrameHeader  mp3;
    } udata;

    for (j=0;j<4;j++)
        udata.raw[j] = buf[j];

    if ((udata.mp3.end_synch != 0x0F) && (udata.mp3.end_synch != 0x0E))
        return FalseWithError("Invalid end_synch byte");

    mp3->sample_hz = 0;
    mp3->bitrate_kbps = 0;
    mp3->duration_ms = 0;
    mp3->frame_length = 0;

    if (udata.mp3.end_synch == 0x0E)
        mp3->mpeg_version = mveVersion2_5;
    else
        mp3->mpeg_version = udata.mp3.id ? mveVersion1_0 : mveVersion2_0;

    mp3->mpeg_layer = MpegLayerArray[udata.mp3.layer];

    if (mp3->mpeg_layer == 0)
    {
        return FalseWithError("Reserved MPEG layer not supported");
    }
    mp3->bitrate_kbps = MpegBitRateArray[mp3->mpeg_version]
                                       [mp3->mpeg_layer - 1]
                                       [udata.mp3.bit_rate];

    if (mp3->bitrate_kbps == 0)
    {
        return FalseWithError("Unsupported bitrate");
    }

    mp3->sample_hz = MpegSampleArray[mp3->mpeg_version]
                                    [udata.mp3.sample_rate];

    if (mp3->sample_hz == 0)
        return FalseWithError("Unsupported sample rate");

    mp3->copyright = (bool)udata.mp3.copyright;
    mp3->original  = (bool)udata.mp3.original;
    mp3->padding   = (bool)udata.mp3.pad;
    mp3->protection= !(bool)udata.mp3.protection;

    // Calculate the frame length
    if (mp3->mpeg_layer == 1)
    {
        mp3->frame_length = (12 * (mp3->bitrate_kbps * 1000) / mp3->sample_hz);

        // Take into account the padding flag
        if (mp3->padding)
            mp3->frame_length++;

        // Each slot is 4 bytes long
        mp3->frame_length *= 4;
    }
    else
    {
        int factor;

        if ((mp3->mpeg_version == mveVersion2_0) && (mp3->mpeg_layer == 3))
            factor = 72;
        else
            factor = 144;

        mp3->frame_length = (factor * (mp3->bitrate_kbps * 1000) / mp3->sample_hz);

        // Take into account the padding flag
        if (mp3->padding)
            mp3->frame_length++;
    }

    // Calculate a CBR total duration in seconds and the frame duration in
    // milliseconds
    if (mp3->bitrate_kbps)
    {
        mp3->duration_ms = (mp3->file_length * 8) / (mp3->bitrate_kbps);
        mp3->frame_duration_ms = (mp3->frame_length * 8) / (mp3->bitrate_kbps);
    }

    return true;
}
///////////////////////////////////////////////////////////////////////////////
long CheckImageForId3V2(const void *vptr)
{
    unsigned char *ptr = (unsigned char *)vptr;

    Mp3Id3v2Header *id3 = (Mp3Id3v2Header *)ptr;

    if ((memcmp(id3->signature, ID3_V2_SIGNATURE, sizeof(id3->signature)) == 0) &&
        ((id3->ver_major & 0x80) == 0)
       )
    {
        // Hmm, this looks like a valid ID3 tag

        // Decode the tag_size field.  This field is encoded to be sync-safe.
        union
        {
            struct
            {
                bits32  zero    :7;
                bits32  one     :7;
                bits32  two     :7;
                bits32  three   :7;
                bits32  padding :4;
            } s;

            bits32 u32;

        } size;

        size.u32 = 0;

        // Copy the values into their decoded bitfields, converting from big
        // to little endian.
        size.s.zero  = id3->tag_size[3];
        size.s.one   = id3->tag_size[2];
        size.s.two   = id3->tag_size[1];
        size.s.three = id3->tag_size[0];

#if 0
        printf("ID3v2 tag size: 0x%08X (%d)\n", size.u32, size.u32);
#endif

        // Return the size of the tag plus the header
        return size.u32 + sizeof(Mp3Id3v2Header);
    }

    return 0;
}
///////////////////////////////////////////////////////////////////////////////
bool GetQuickInfoFromMp3Image(long flength, const void *vptr, Mp3Info *mp3)
{
    unsigned char *ptr = (unsigned char *)vptr;
    long pos, new_pos;
    int valid_headers;
    unsigned char buf[4];

    if (vptr == NULL)
        return false;

    memset(mp3, 0, sizeof(*mp3));

    mp3->file_length = flength;

    if (mp3->file_length < 5)
    {
        return false;
    }

    valid_headers = 0;

    // Check for an initial id3v2 tag (and skip over it)
    if ((pos = CheckImageForId3V2(vptr)) != 0)
    {
        mp3->id3v2_length = pos;
    }

    new_pos = pos;

    // Try to find a valid frame header.  Keep trying until the desired number
    // of consecutive frames has been found

    while(pos < flength && valid_headers < consecutive_valid_headers_needed)
    {
        // Don't go farther than 1 Meg into the file to look for a header
        if (pos > (new_pos + 1024 * 1024))
            return false;

        memset(buf, 0, sizeof(buf));

        if ((flength - pos) >= (long)sizeof(buf))
        {
            memcpy(buf, &ptr[pos], sizeof(buf));

            if (buf[0] == 0xFF)
            {
                if (ReadMp3FrameHeader((char*)&buf[0], mp3))
                {
                    valid_headers++;

                    mp3->frame_offset = pos;

                    if (valid_headers == 1)
                        mp3->first_frame_offset = pos;

                    pos += mp3->frame_length;
                    continue;
                }
                else
                {
                    valid_headers = 0;
                    pos++;
                }
            }
            else
            {
                valid_headers = 0;
                pos++;
            }
        }
        else
            return false;

    }

    // If we exited the loop before reaching eof, then we must have found a
    // valid header

    if (pos < mp3->file_length)
        return true;


    // Otherwise, we must not have.
    return false;
}
///////////////////////////////////////////////////////////////////////////////
bool GetQuickInfoFromMp3File(const char *fname, Mp3Info *mp3)
{
    FILE *fp;
    long pos, new_pos;
    int valid_headers;
    unsigned char buf[4];

    if ((fp = fopen(fname, "rb")) == NULL)
        return false;

    memset(mp3, 0, sizeof(*mp3));

    mp3->file_length = GetFileLengthFromDescriptor(fileno(fp));
    if (mp3->file_length < 5)
    {
        fclose(fp);
        return false;
    }

    valid_headers = 0;

    char header[10] = {0};
    fread(header, 1, sizeof(header), fp);

    // Check for an initial id3v2 tag (and skip over it)
    if ((pos = CheckImageForId3V2(header)) != 0)
    {
        mp3->id3v2_length = pos;
    }

    new_pos = pos;

    // Try to find a valid frame header.  Keep trying until the desired number
    // of consecutive frames has been found

    while(!feof(fp) && valid_headers < consecutive_valid_headers_needed)
    {
        // Don't go farther than 1 Meg into the file to look for a header
        if (pos > (new_pos + 1024 * 1024))
        {
            fclose(fp);
            return false;
        }

        fseek(fp, pos, SEEK_SET);

        memset(buf, 0, sizeof(buf));

        if (fread(buf, 1, sizeof(buf), fp) == sizeof(buf))
        {
            if (buf[0] == 0xFF)
            {
                if (ReadMp3FrameHeader((char*)&buf[0], mp3))
                {
                    valid_headers++;

                    mp3->frame_offset = pos;

                    if (valid_headers == 1)
                        mp3->first_frame_offset = pos;

                    pos += mp3->frame_length;
                    continue;
                }
                else
                {
                    valid_headers = 0;
                    pos++;
                }
            }
            else
            {
                valid_headers = 0;
                pos++;
            }
        }
        else
        {
            fclose(fp);
            return false;
        }
    }

    fclose(fp);

    // If we exited the loop before reaching eof, then we must have found a
    // valid header

    if (pos < mp3->file_length)
        return true;


    // Otherwise, we must not have.
    return false;
}
///////////////////////////////////////////////////////////////////////////////
bool GetFullInfoFromMp3Image(Mp3Info *pinfo, long flength, const void *vptr, void *user, Mp3Callback callback)
{
    unsigned char *ptr = (unsigned char *)vptr;
    long pos;
    unsigned char buf[4];
    Mp3Info mp3obj;
    Mp3Info *mp3 = ( pinfo ? pinfo : &mp3obj);

    unsigned total_frames = 0;

    if (ptr == NULL || flength < 5)
        return false;

    memset(mp3, 0, sizeof(*mp3));
    mp3->first_frame_offset = -1;

    // Check for an initial id3v2 tag (and skip over it)
    if ((pos = CheckImageForId3V2(vptr)) != 0)
    {
        mp3->id3v2_length = pos;
    }

    // Walk through all frames.

    while(pos < flength)
    {
        memset(buf, 0, sizeof(buf));

        if ((flength - pos) >= (long)sizeof(buf))
        {
            memcpy(buf, &ptr[pos], sizeof(buf));

            if (buf[0] == 0xFF)
            {
                mp3->file_length = flength;

                if (ReadMp3FrameHeader((char *)&buf[0], mp3))
                {
                    total_frames++;

                    mp3->frame_offset = pos;

                    if (mp3->first_frame_offset == -1)
                        mp3->first_frame_offset = pos;

                    if (callback)
                    {
                        if (callback(user, mp3) == false)
                            return true;
                    }

                    pos += mp3->frame_length;
                    continue;
                }
                else
                {
                    pos++;
                }
            }
            else
            {
                pos++;
            }
        }
        else
        {
            // We've just about hit the end of the file, no more frames could
            // possibly follow, since we're within the frame header size
            break;
        }
    }

    // We've reached the end of the file

    return (total_frames >= consecutive_valid_headers_needed);

}
///////////////////////////////////////////////////////////////////////////////
bool GetFullInfoFromMp3File(Mp3Info *pinfo, const char *fname, void * user, Mp3Callback callback)
{
    FILE *fp;
    long pos;
    unsigned char buf[4];
    Mp3Info mp3obj;
    Mp3Info *mp3 = ( pinfo ? pinfo : &mp3obj);
    long flength;
    unsigned total_frames = 0;

    if ((fp = fopen(fname, "rb")) == NULL)
        return false;

    flength = GetFileLengthFromDescriptor(fileno(fp));
    if (flength < 5)
    {
        fclose(fp);
        return false;
    }

    memset(mp3, 0, sizeof(*mp3));
    mp3->first_frame_offset = -1;

    char header[10] = {0};
    fread(header, 1, sizeof(header), fp);

    // Check for an initial id3v2 tag (and skip over it)
    if ((pos = CheckImageForId3V2(header)) != 0)
    {
        mp3->id3v2_length = pos;
    }

    // Walk through all frames.

    while(!feof(fp))
    {
        fseek(fp, pos, SEEK_SET);

        memset(buf, 0, sizeof(buf));

        if (fread(buf, 1, sizeof(buf), fp) == sizeof(buf))
        {
            if (buf[0] == 0xFF)
            {
                mp3->file_length = flength;

                if (ReadMp3FrameHeader((char *)&buf[0], mp3))
                {
                    total_frames = 0;

                    mp3->frame_offset = pos;

                    if (mp3->first_frame_offset == -1)
                        mp3->first_frame_offset = pos;

                    if (callback)
                    {
                        if (callback(user, mp3) == false)
                            return true;
                    }

                    pos += mp3->frame_length;
                    continue;
                }
                else
                {
                    pos++;
                }
            }
            else
            {
                pos++;
            }
        }
        else
        {
            // We've just about hit the end of the file, no more frames could
            // possibly follow, since we're within the frame header size
            break;
        }
    }

    // We've reached the end of the file

    return (total_frames >= consecutive_valid_headers_needed);
}
///////////////////////////////////////////////////////////////////////////////

#ifdef USE_MAIN

static char *crack_time(int s)
{
    static char buf[10];

    int hr, min, sec;

    hr = s / (60 * 60);

    s = s - hr * (60 * 60);

    min = s / 60;

    s = s - min * 60;

    sec = s;

    if (hr)
        sprintf(buf, "%02d:%02d:%02d", hr, min, sec);
    else
        sprintf(buf, "%02d:%02d", min, sec);

    return buf;
}

static int vbr_average = 0;
static int total_frames = 0;
static int total_duration_ms = 0;
static bool is_vbr = false;

static bool ForEachFrame(void * /* user */, Mp3Info *mp3)
{
    static int last_bitrate = -1;

    if (last_bitrate == -1)
    {
        last_bitrate = mp3->bitrate_kbps;
    }
    else
    if (mp3->bitrate_kbps != last_bitrate)
    {
        is_vbr = true;
        mp3->is_vbr = true;
    }

    vbr_average += mp3->bitrate_kbps;
    total_frames++;
    total_duration_ms += mp3->frame_duration_ms;
#ifdef DUMP_ALL_FRAMES
    printf("header @ 0x%08X ", mp3->frame_offset);
    printf("{frmlen:%04X}", mp3->frame_length);
    printf("(%3dKbps) ", mp3->bitrate_kbps);
    printf("<%5dHz>\n", mp3->sample_hz);
#endif

    return true;
}

#pragma argsused
int main (int argc, char *argv[])
{
    Mp3Info mp3;
    float ave;
    int total_duration;

    assert(sizeof(Mp3FrameHeader) == 4);

    if (argc < 2)
    {
        printf ("Usage: mp3hdr.exe mp3filename\n");
        return 0;
    }

    if (GetQuickInfoFromMp3File(argv[1], &mp3) == false)
    {
        printf("Error locating mp3 header info\n");
        return 1;
    }

    GetFullInfoFromMp3File(NULL, argv[1], NULL, ForEachFrame);

    ave = (vbr_average * 1.0) / total_frames;
    total_duration = total_duration_ms / 1000;

    printf("\n");
    printf("file:            %s\n", argv[1]);
    printf("size:            %d bytes\n", mp3.file_length);
    if (mp3.id3v2_length)
        printf("ID3 v2 length:   %d bytes\n", mp3.id3v2_length);
    printf("header found at: %d bytes\n", mp3.first_frame_offset);
    printf("Total frames:    %d\n", total_frames);
    printf("MPEG Version     MPEG %s layer %d\n", mp3.GetMpegVersionString(), mp3.mpeg_layer);
    printf("Copyrighted:     %d\n", mp3.copyright);
    printf("Original:        %d\n", mp3.original);
    printf("CRCs:            %d\n", mp3.protection);
    printf("sample freq:     %dHz\n", mp3.sample_hz);
    if (is_vbr == false)
    {
        printf("duration:        %s (%d seconds)\n", crack_time(mp3.duration_ms / 1000), mp3.duration_ms / 1000);
        printf("bitrate:         %dKbps\n", mp3.bitrate_kbps);
    }
    else
    {
        printf("duration:        %s (%d seconds)\n", crack_time(total_duration), total_duration);
        printf("VBR ave. bitrate %.1fKbps\n", ave);
    }
    return 0;
}
#endif
